# trc
# trace read compression
# author: roye rozov, 2013

import sys, os, argparse
import config
import scratch
from Bio import SeqIO
from scipy.optimize import minimize
from scipy.misc import comb
from bitarray import bitarray
from binascii import unhexlify


def cost(r, err_rate, Gen_len=6e7, n=6e6, end_len=100, read_len=100):
	Q = 2*Gen_len*(3*end_len+1)
	c=.6185
	# constants assuming 0.003 base error rate
	# read length 100
	one_err_rate = comb(read_len, 1) * (1-err_rate)**(read_len - 1) * \
		err_rate #0.247 #0.223
	two_err_rate = comb(read_len, 2) * (1-err_rate)**(read_len - 2) * \
		err_rate**2 #0.043 #0.033
	# bf_cost = (1+((2*Q*c**(2*r))/n))* r
	# expl_cost = ((1-(float(end_len)/read_len)) * one_err_rate + 2*two_err_rate + c**(2*r)) * 2 * read_len

	p_cost = (1+(Q*c**r)/n)* (r / (1 - c**r))
	n_cost = 0 #((1-(float(end_len)/read_len)) * one_err_rate + 2*two_err_rate) * read_len
	# print "bf cost is %f, expl cost is %f" % (bf_cost, expl_cost)
	# return bf_cost + expl_cost
	if (end_len != 0):
		print "p cost is %.3f, n cost is %.3f" % (p_cost, n_cost)
		return p_cost + n_cost
	else:	
		print p_cost
		return p_cost

def init_position_vals(read_len):
	config.READ_LEN = read_len
	config.READ_POSITION_EXPS = [(4**(read_len-read_pos-1)) for read_pos in range(read_len)]
	config.mask1_vals = [(4**(read_len-read_pos-1)) for read_pos in range(read_len-1,-1,-1)]
	config.mask2_vals = [(2 * 4**(read_len-read_pos-1)) for read_pos in range(read_len-1,-1,-1)]
	return

# http://stackoverflow.com/a/14527004
def long_to_bytes (val, endianness='big'):
	"""
	Use :ref:`string formatting` and :func:`~binascii.unhexlify` to
	convert ``val``, a :func:`long`, to a byte :func:`str`.

	:param long val: The value to pack

	:param str endianness: The endianness of the result. ``'big'`` for
	big-endian, ``'little'`` for little-endian.

	If you want byte- and word-ordering to differ, you're on your own.

	Using :ref:`string formatting` lets us use Python's C innards.
	"""

	# one (1) hex digit per four (4) bits
	width = val.bit_length()

	# unhexlify wants an even multiple of eight (8) bits, but we don't
	# want more digits than we need (hence the ternary-ish 'or')
	width += 8 - ((width % 8) or 8)

	# format width specifier: four (4) bits per hex digit
	fmt = '%%0%dx' % (width // 4)

	# prepend zero (0) to the width, to zero-pad the output
	s = unhexlify(fmt % val)

	if endianness == 'little':
		# see http://stackoverflow.com/a/931095/309233
		s = s[::-1]
	return s


parser = argparse.ArgumentParser()
# parses parameters based on command entered
subparsers = parser.add_subparsers(help='commands', dest='command')

parser_rload = subparsers.add_parser('rec_load_bf', help='recursively loads reads & next stage FPs into Bloom Filters')
parser_rload.add_argument('reads', action='store', help='2 bit packed reads file')
parser_rload.add_argument('genome', action='store', help='2 bit packed reference genome fasta file')
parser_rload.add_argument('coverage', action='store', help='mean coverage level', type=int)
parser_rload.add_argument('-err_rate', action='store', help='per base error rate', type=float, default=0.0)
parser_rload.add_argument('-read_len', action='store', help='read length --optional, default is 100',
						type=int, default=100)
parser_rload.add_argument('-e', '-end_len', action='store', help='length (<= read_len) at end to check all substitutions',
						type=int, default=0)
parser_rload.add_argument('-i', '-iter', action='store', help='number of bloom filters wanted',
						type=int, default=4)


parser_decode = subparsers.add_parser('decode', help='decodes')
parser_decode.add_argument('genome', action='store', help='2 bit packed reference genome fasta file')
parser_decode.add_argument('prefix', action='store', help='prefix common to bf files')
parser_decode.add_argument('fp_i_file', action='store', help='filename of ancillary FP set')

parser_decode.add_argument('-read_len', action='store', help='read length --optional, default is 100',
						type=int, default=100)
parser_decode.add_argument('-e', '-end_len', action='store', help='length (<= read_len) at end to check all substitutions',
						type=int, default=0)
parser_decode.add_argument('-i', '-iter', action='store', help='number of bloom filters in cascade',
						type=int, default=4)



arg_set = parser.parse_args()
print arg_set

### to rec. load bf & FPs 
if (arg_set.command=='rec_load_bf'):
	genome_len = os.stat(arg_set.genome).st_size  * 4 # approx. length from file size  (*4 b/c using packed)
	numreads = genome_len * arg_set.coverage / arg_set.read_len
	print "genome_len is %d, numreads is %d" % (genome_len, numreads)
	vals = minimize(cost, x0=5, args=(arg_set.err_rate, genome_len, numreads, arg_set.e, arg_set.read_len), method='L-BFGS-B')
	# vals = minimize(cost, 1, args=(genome_len, numreads, arg_set.e, arg_set.read_len))
	r = vals.x
	exp_cost = vals.fun
	c = 0.6185
	f = c**r
	print "expected cost per unique read is %f, f is %.3f" % (exp_cost, f)
	init_position_vals(arg_set.read_len)
	scratch.test_init()
	(FP_i, FNs) = scratch.rec_get_bf(f, arg_set.e, arg_set.reads, arg_set.genome, arg_set.i)

	# write out FP_i
	d = {'A':bitarray('00'), 'C':bitarray('01'), 'G':bitarray('10'), 'T':bitarray('11')}
	fp_filename = arg_set.reads + ".e" +str(arg_set.e) +".leftover.FPs"
	FP_OUT = open(fp_filename, "wb")
	FP_cnt = 0
	for s in FP_i:
		if (s == 0):
			bytes = '\0x0' ## I think this might not be right - don't get 00000000
		else:
			bytes = long_to_bytes(s)
		b = bitarray()
		b.frombytes(bytes)
		dna = ''.join(b.decode(d))
		# to avoid trunctation due to leading 0s
		dna = (arg_set.read_len - len(dna))*'A' + dna
		print >>FP_OUT, dna
	FP_OUT.close()

	# convert FNs back to sequences, write out
	fn_filename = arg_set.reads + ".e" +str(arg_set.e) +".leftover.FNs"
	FN_OUT = open(fn_filename, "wb")
	FN_cnt = 0
	for t in FNs:
		if (t == 0):
			bytes = '\0x0' ## I think this might not be right - don't get 00000000
		else:
			bytes = long_to_bytes(t)
		b = bitarray()
		b.frombytes(bytes)
		dna = ''.join(b.decode(d))
		# to avoid trunctation due to leading 0s
		dna = (arg_set.read_len - len(dna))*'A' + dna

		for mult in range(FNs[t]):
			print >>FN_OUT, dna
			FN_cnt += 1
	FN_OUT.close()
	print "%d FPs, %d FNs, %d total " % (len(FP_i), FN_cnt, len(FP_i) + FN_cnt)

# decompress reads given G, B1-B4, FP_4
# FN/FP_4 are decompressed separately
# output 2 files - R_rc reads (from BF rejections)
if (arg_set.command=='decode'):
	init_position_vals(arg_set.read_len)
	d = {'A':bitarray('00'), 'C':bitarray('01'), 'G':bitarray('10'), 'T':bitarray('11')}
	FPs = []
	left_over_seqs = []
	R_rc_seqs = []
	(R_rc, left_over) = scratch.decode(arg_set.e, arg_set.prefix, arg_set.genome, arg_set.i)

	for s in left_over:
		if (s == 0):
			bytes = '\0x0' ## I think this might not be right - don't get 00000000
		else:
			bytes = long_to_bytes(s)
		b = bitarray()
		b.frombytes(bytes)
		dna = ''.join(b.decode(d))
		# to avoid trunctation due to leading 0s
		dna = (arg_set.read_len - len(dna))*'A' + dna
		left_over_seqs.append(dna)

	for t in R_rc:
		if (t == 0):
			bytes = '\0x0' ## I think this might not be right - don't get 00000000
		else:
			bytes = long_to_bytes(t)
		b = bitarray()
		b.frombytes(bytes)
		dna = ''.join(b.decode(d))
		# to avoid trunctation due to leading 0s
		dna = (arg_set.read_len - len(dna))*'A' + dna
		R_rc_seqs.append(dna)

	R_rc_seqs = set(R_rc_seqs)
	left_over_seqs = set(left_over_seqs)
	print "%d reads in R_rc, %d reads in left_over_seqs before merge" % (len(R_rc), len(left_over_seqs))


	FP_file = open(arg_set.fp_i_file, "r")
	for s in SeqIO.parse(FP_file, "fastq"):
		FPs.append(s.seq.strip())
	FP_file.close()

	FPs = set(FPs)
	print "%d FPs before intersect" % len(FPs)
	rest = FPs.intersection(left_over_seqs)
	print "size of rest is %d" % len(rest)
	
	# print "size of union b/w R_rc and FPs is %d" % len(R_rc_seqs.union(FPs))
	R_rc_seqs = R_rc_seqs.union(rest)
	print "%d reads in R_rc union with rest" % len(R_rc)



